//#include <exception>

//	Copyright (C) by Attila Lendvai
//	You are free to use this file in anything you want as long as it's clear
//	that it's not your work. Use at your own risk, no warranties!

// BUGS:
// horizontal scroller does not work

#include <Autolock.h>
#include <MessageQueue.h>
#include <Roster.h>
#include <MenuBar.h>
#include <PopUpMenu.h>
#include <MenuItem.h>
#include <Path.h>
#include <Directory.h>
#include <Debug.h>
#include <File.h>
#include <time.h>

#include "Preferences.h"
#include "Misc.h"
#include "DebugStream.h"

#if USE_DEBUGSTREAM

using namespace std;
using namespace misc;
using namespace preferences;
using namespace debug;
using namespace debug_internal;

DebugManager DebugManager :: sManager;
int32 DebugManager :: sDebugStream_TLS = 0;

static const char debug_dir_name[] = "Debug_Output";

DebugManager :: DebugManager() : mClearLogsAtStartup(true), mLogToFileByDefault(true), mLocker("DebugManager :: mLocker"),
								 mWindow(0), mInitialised(0) {
}

void DebugManager :: Initialise() {
	if (atomic_or(&mInitialised, 1) == 0) {
		sDebugStream_TLS = tls_allocate();
		{
			Preferences *prefs = gGetUserPreferences();
			Autolock<Preferences> ll(prefs);
			
			bool b;
			if ( prefs -> FindBool("dstream:ClearLogsAtStartup", &b) == B_OK )
				mClearLogsAtStartup = b;
			if ( prefs -> FindBool("dstream:LogToFileByDefault", &b) == B_OK )
				mLogToFileByDefault = b;
		}
		
		// XXX some more tolerant code with exceptions
		thread_info ti;
		get_thread_info(find_thread(0), &ti);
		app_info ai;
		be_roster -> GetRunningAppInfo(ti.team, &ai);
		BEntry appentry(&ai.ref), appdir_entry;
		appentry.GetParent(&appdir_entry);
		BDirectory appdir(&appdir_entry);
		BDirectory dir(&appdir, debug_dir_name), dirtmp;
		BEntry entry;
		switch (dir.InitCheck()) {
			case B_OK:
				entry.SetTo(&appdir, debug_dir_name);
				entry.GetRef(&mLogDir);
				break;

			case B_ENTRY_NOT_FOUND: {
				dir.SetTo(&appdir_entry);
				if (dir.CreateDirectory(debug_dir_name, &dirtmp) != B_OK) {
					return;	// +++ exception
				}
				entry.SetTo(&dir, debug_dir_name);
				entry.GetRef(&mLogDir);
				dir.SetTo(&entry);
				break;
			}
			default:
				// +++ exception
				return;
				break;
		}
		if (mClearLogsAtStartup) {
			dir.Rewind();
			while (dir.GetNextEntry(&entry) == B_OK) {
				entry.Remove();
			}
		}
	}
}

DebugManager :: ~DebugManager() {
	if ( mWindow && mWindow -> Lock() ) {
		mWindow -> Quit();
		mWindow = 0;
	}
	
	stream_list_t :: const_iterator pos = mStreams.begin(), end = mStreams.end();
	while (pos != end) {
		delete *pos;
		++pos;
	}
	
	Preferences *prefs = gGetUserPreferences();
	Autolock<Preferences> ll(prefs);
	
	prefs -> SetBool("dstream:ClearLogsAtStartup", mClearLogsAtStartup);
	prefs -> SetBool("dstream:LogToFileByDefault", mLogToFileByDefault);
}

DebugStream &DebugManager :: GetStream() {
	Initialise();

	DebugStream *stream = (DebugStream *)tls_get(sDebugStream_TLS);
	if (stream == 0) {
		BAutolock l(Locker());
		
		stream = new DebugStream( find_thread(0) );
		mStreams.push_back(stream);
//		bool locked = false;
		if ( mWindow ) { // &&  (locked = mWindow -> Lock()) == true ) {
			BMessage msg( M_NEW_STREAM );
			msg.AddPointer( "Stream", stream );
			mWindow -> PostMessage( &msg );
//			if (locked)
//				mWindow -> Unlock();
		}
	}
	return reinterpret_cast<DebugStream &>(*stream);
}


DebugStream :: DebugStream(thread_id id, const char *fname) : ostream(&mBuffer), ios(0), mThreadID(id),
													mWindow(0), mTimeStampState(true), mBuffer(this)  {
	thread_info ti;
	get_thread_info(id, &ti);
	mName = ti.name;
	
	BPath path(gGetDebugManager().LogDir());
	if (path.InitCheck() == B_OK) {
		BString str(path.Path());
		str += "/";
		str += mName;
		mBuffer.SetFile(str.String());
	} // +++ exception ?
	
	SetFileLoggingEnabled(gGetDebugManager().mLogToFileByDefault);
}

void DebugStream :: OpenWindow() {
	bool locked = false;
	if ( mWindow == 0 || (locked = mWindow -> Lock()) == false ) {
		Window *win = new Window();
		Attach( win );
		win -> Attach(this);
	}
	mWindow -> Minimize(false);
	mWindow -> Show();
	if (locked)
		mWindow -> Unlock();
}

DebugStream :: ~DebugStream() {
	bool locked = false;
	if ( mWindow && (locked = mWindow -> Lock()) == true ) {
		mWindow -> Detach(this);
		mWindow -> Unlock();
	}
}

void DebugStream :: Attach(Window *inWin)	{
	ASSERT_WITH_MESSAGE(mWindow == 0, "debug::DebugStream::Attach called while already attached");
	mWindow = inWin;
}

void DebugStream :: Detach(Window *inWin)	{
	ASSERT_WITH_MESSAGE(mWindow == inWin, "debug::DebugStream::Detach called with a different window");
	mWindow = 0;
}







void DebugManager :: OpenMainWindow() {
	sManager.OpenMainWindowMember();
}

void DebugManager :: OpenMainWindowMember() {
	mWindow = new MainWindow( this );
	mWindow -> Show();
}





Buffer :: Buffer(DebugStream *inOwner) : mOwner(inOwner), mPreviousFlushAt(0),
										 mFileLoggingState(false), mFlushed(true) {
	open("/dev/null", ios::out | ios::app);
	mPreviousFlushAt = pptr();
}

//Buffer :: ~Buffer() {
//}

void Buffer :: ReopenFile() {
	close();
	if (mFileLoggingState)
		open(mFileName.String(), ios::out | ios::app);
	else
		open("/dev/null", ios::out | ios::app);
}

int Buffer :: doallocate() {
	int i = inherited :: doallocate();
	return i;
}

void Buffer :: SetFileLoggingEnabled(bool b) {
	if (mFileLoggingState == b)
		return;
		
	mFileLoggingState = b;
	
	if (b == true)
		ReopenFile();
		
	char buf1[128], buf2[128];
	CreateTimeStamp(buf1);
//	sprintf(buf2, "%sFile logging %s.\n", buf1, (b) ? "enabled" : "disabled");
	sprintf(buf2, "File logging %s.\n", (b) ? "enabled" : "disabled");
	xsputn(buf2, strlen(buf2));
	sync();
	
//	Window *win = mOwner -> mWindow;
//	if ( win != 0 ) {
//		BAutolock l( win );
//		if ( l.IsLocked() ) {
//			win -> Append( buf2, strlen(buf2) );
//		}
//	}
	if (b == false)
		ReopenFile();
}

int Buffer :: sync() {
	mFlushed = true;
	
//	mBuffer.append(pbase(), out_waiting());
		
	Window *win = mOwner -> mWindow;
	if ( win != 0 ) {
		BAutolock l( win );
		if ( l.IsLocked() ) {
//cerr << "sync: " << pbase() << "; count: " << out_waiting() << endl;	// XXX
			char *pos = pptr();
			win -> Append( mPreviousFlushAt, pos - mPreviousFlushAt );
			mPreviousFlushAt = pos;
		} else {
			// Clear Window if it's gone...
			if ( win  -> Lock() == false)
				mOwner -> mWindow = 0;
			else
				win -> Unlock();
		}
	}
	
//	int rc = 0;
	return inherited :: sync();
//	return 0;
}

//void Buffer :: close() {
//	inherited :: close();
//}

streamsize Buffer :: xsputn(const char* s, streamsize n) {
//cerr << "xsputn: " << s << "size: " << n << endl;	// XXX
	if (mFlushed) {
		mFlushed = false;
		mPreviousFlushAt = pptr();
		
		if (mOwner -> IsTimeStampEnabled()) {
			char buf[128];
			CreateTimeStamp(buf);
			
//			Window *win = mOwner -> GetWindow();
//			if ( win  &&  win -> Lock() ) {
//				TextView *tv = win -> mTextView;
//				BFont font;
//				rgb_color color = {0, 255, 0};
//				tv -> SetFontAndColor(tv -> TextLength(), tv -> TextLength(), &font, 0, &color);
//				
//				inherited :: xsputn(buf, strlen(buf));
//				sync();
//				mFlushed = false;
//				
//				color.red = 0;
//				color.green = 0;
//				color.blue = 0;
//				tv -> SetFontAndColor(tv -> TextLength(), tv -> TextLength(), &font, 0, &color);
//				win -> Unlock();
//			} else {
				inherited :: xsputn(buf, strlen(buf));
//			}
		}
	}

	return inherited :: xsputn(s, n);
}

//int	Buffer :: underflow() {
//	cerr << "underflow" << endl;
//	return inherited :: underflow();
//}
//

void Buffer :: CreateTimeStamp(char *buf) {
	struct tm tme;
	time_t now = time(&now);
	localtime_r(&now, &tme);
		
	sprintf(buf, "%.2ld:%.2ld:%.2ld \t", (int32)tme.tm_hour, (int32)tme.tm_min, (int32)tme.tm_sec);
}

int Buffer :: overflow(int ch)	{
	bool stamp = false;
	char buf[128];
	
	if (mFlushed) {
		mFlushed = false;
		
		if (mOwner -> IsTimeStampEnabled()) {
			CreateTimeStamp(buf);
			stamp = true;
//			Window *win = mOwner -> GetWindow();
//			if ( win  &&  win -> Lock() ) {
//				TextView *tv = win -> mTextView;
//				BFont font;
//				rgb_color color = {0, 255, 0};
//				tv -> SetFontAndColor(tv -> TextLength(), tv -> TextLength(), &font, 0, &color);
//				
//				inherited :: xsputn(buf, strlen(buf));	// XXX
//				sync();
//				mFlushed = false;
//				
//				color.red = 0;
//				color.green = 0;
//				color.blue = 0;
//				tv -> SetFontAndColor(tv -> TextLength(), tv -> TextLength(), &font, 0, &color);
//				win -> Unlock();
//			} else {
				sys_write(buf, strlen(buf));
//			}
		}
	}
	Window *win = mOwner -> mWindow;
	if ( win != 0 ) {
		BAutolock l( win );
		if ( l.IsLocked() ) {
			if (stamp)
				win -> Append( buf, strlen(buf) );
			int size = epptr() - pbase();
			if (size)
				win -> Append( pbase(), size);	// XXX ok?
		} else {
			// Clear Window if it's gone...
			if ( win  -> Lock() == false)
				mOwner -> mWindow = 0;
			else
				win -> Unlock();
		}
	}
	int rc = inherited :: overflow(ch);
	mPreviousFlushAt = pptr();
	return rc;
}

//streamsize Buffer :: sys_read(char* buf, streamsize size) {
//	cerr << "sys_read: size: " << size << endl;
//	return inherited :: sys_read(buf, size);
//}
//
//streampos Buffer :: sys_seek(streamoff off, _seek_dir dir) {
//	cerr << "sys_seek: " << off << dir << endl;
//	return inherited :: sys_seek(off, dir);
//}
//
//streamsize Buffer :: sys_write(const char* buf, streamsize size) {
//	cerr << "sys_write: " << buf << size << endl;
//	return inherited :: sys_write(buf, size);
//}

int32 Window :: sInstances = 0;
BPoint Window :: sLeftTop(20, 20);

Window :: Window() : BWindow(BRect(20, 20, 120, 120), "Debug", B_DOCUMENT_WINDOW, 0),
					 mStream(0) {
					 
	if ( atomic_add(&sInstances, 1) == 0 ) {
		MainWindow *win = DebugManager :: GetMainWindow();
		{
			BAutolock l(win);
			if (l.IsLocked()) {
				sLeftTop = win -> Frame().LeftTop();
			}
		}
	}
	
	sLeftTop.x += 10;
	sLeftTop.y += 10;
	MoveTo(sLeftTop);
	
	SetPulseRate(500000);
	mTextView = new TextView();
	mTextView -> ResizeBy(-B_V_SCROLL_BAR_WIDTH, -B_H_SCROLL_BAR_HEIGHT);
	
	AddChild(new BScrollView("ScrollView", mTextView, B_FOLLOW_ALL, 0, true, true));
	MainWindow *win = gGetDebugManager().GetMainWindow();
	if (win) {
		BRect rect = win -> Bounds();
		ResizeTo(rect.Width(), rect.Height());
	} else {
		ResizeTo(200, 300);
	}
	
	mTextView -> Pulse();
}


Window :: ~Window() {
	atomic_add(&sInstances, -1);
	if (mStream) {
		mStream -> Detach(this);
		MainWindow *win = DebugManager :: GetMainWindow();
		if (win)
			win -> PostMessage(M_WINDOW_GONE);
	}
}
					
void Window :: Detach(DebugStream *stream) {
	ASSERT_WITH_MESSAGE(mStream == stream, "debug::Window::Detach called with a non-attached stream");
	mStream = 0;
}

void Window :: Attach(DebugStream *stream) {
	ASSERT_WITH_MESSAGE(mStream == 0, "debug::Window::Attach called while already attached");
	mStream = stream;
	
//	mTextView -> SetText(mStream -> mBuffer.mBuffer.data(), mStream -> mBuffer.mBuffer.size());
	
	BFile file(mStream -> FileName() -> String(), B_READ_ONLY);
	if (file.InitCheck() == B_OK) {
		off_t size;
		file.GetSize(&size);
		mTextView -> SetText(&file, 0, size);
	}
}

void Window :: Append(const char *inStr, int inSize) {
	mTextView -> Insert(mTextView -> TextLength(), inStr, inSize);
}

//void Window :: FrameResized(float newWidth, float newHeight) {
//	BMessageQueue *queue = MessageQueue();
//	BMessage *msg;
//	
//	Autolock<BMessageQueue> l(queue);
//	while ( (msg = queue -> FindMessage(B_WINDOW_RESIZED, 0)) != 0 )
//		queue -> RemoveMessage(msg);
//}



TextView :: TextView() : BTextView(BRect(0, 0, 100, 100), "TextView", BRect(0, 0, 100, 100), B_FOLLOW_ALL, B_WILL_DRAW | B_NAVIGABLE | B_FRAME_EVENTS),
						 mNeedSetTextRect(true) {
//Insert("test123\n", 65500);	// XXX
	MakeEditable(false);
	SetWordWrap(false);
}

void TextView :: Pulse() {
	if (mNeedSetTextRect) {
		SetTextRect(Bounds());
		mNeedSetTextRect = false;
	}
	
	inherited :: Pulse();
}

void TextView :: FrameResized(float newWidth, float newHeight) {
	mNeedSetTextRect = true;
}



StreamListItem :: StreamListItem( debug :: DebugStream *inStream ) : BStringItem(""),
											mStream(inStream) {
	thread_info ti;
	get_thread_info( mStream -> ThreadID(), &ti);
	SetText( strlen(ti.name) == 0 ? "Unnamed thread" : ti.name );
}

void StreamListItem :: DrawItem(BView *owner, BRect itemRect, bool drawEverything) {
	rgb_color color;
	if (IsSelected()) {
		color = tint_color(owner -> ViewColor(), B_HIGHLIGHT_BACKGROUND_TINT);
		owner -> SetHighColor(color);
		owner -> SetLowColor(color);
		owner -> FillRect(itemRect);
	} else if (drawEverything) {
		color = owner -> ViewColor();
		owner -> SetHighColor(color);
		owner -> SetLowColor(color);
		owner -> FillRect(itemRect);
	} else {
		owner -> SetLowColor(owner -> ViewColor());
	}
	
	owner -> SetDrawingMode(B_OP_COPY);
	if (Stream() -> HasWindow()  ||  Stream() -> IsFileLoggingEnabled()) {
		owner -> SetHighColor(0, 0, 0);
	} else {
		memset(&color, 0, sizeof(rgb_color));
		owner -> SetHighColor(tint_color(color, B_LIGHTEN_1_TINT));
	}
	BFont font;
	font_height height;
	owner -> GetFont(&font);
	font.GetHeight(&height);
	BPoint point = itemRect.LeftBottom();
	point.y -= height.descent;
	owner -> DrawString(Text(), point);
}


StreamListView :: StreamListView(DebugManager *inManager) :
					BListView(BRect(5, 5, 95, 95), "ListView", B_MULTIPLE_SELECTION_LIST, B_FOLLOW_ALL),
					mManager(inManager) {
					
	DebugManager :: stream_list_t :: const_iterator
		pos = mManager -> StreamList() -> begin(),
		end = mManager -> StreamList() -> end();
	while ( pos != end ) {
		AddItem(new StreamListItem( *pos ));
		pos++;
	}
	
	SetInvocationMessage( new BMessage(M_ITEMS_CLICKED) );
}

void StreamListView :: MessageReceived(BMessage *inMsg) {
	switch ( inMsg -> what ) {
		case M_ITEMS_CLICKED: {
			BAutolock l(gGetDebugManager().Locker());
			
			int32 index;
			for ( int i = 0;  inMsg -> FindInt32("index", i, &index) == B_OK;  i++ ) {
				DebugStream *stream = assert_cast<StreamListItem *>(ItemAt(index)) -> Stream();
				bool has = stream -> HasWindow();
				stream -> OpenWindow();
				if (has != stream -> HasWindow())
					InvalidateItem(index);
			}
			break;
		}
		
		case M_OPEN_WINDOW: {
			BAutolock l(gGetDebugManager().Locker());
			
			void *ptr;
			if (inMsg -> FindPointer("sitem", &ptr) == B_OK) {
				StreamListItem *sitem = static_cast<StreamListItem *>(ptr);
				DebugStream *stream = sitem -> Stream();
				bool has = stream -> HasWindow();
				stream -> OpenWindow();
				if (has != stream -> HasWindow())
					InvalidateItem(IndexOf(sitem));
			}
			break;
		}
		
		case M_LOG_SELECTED_TO_FILE: {
			BAutolock l(gGetDebugManager().Locker());
			
			int count = CountItems();
			for ( int i = 0;  i < count;  i++ ) {
				if (IsItemSelected(i)) {
					DebugStream *stream = assert_cast<StreamListItem *>(ItemAt(i)) -> Stream();
					bool b = stream -> IsFileLoggingEnabled();
					stream -> SetFileLoggingEnabled(true);
					if (b != stream -> IsFileLoggingEnabled())
						InvalidateItem(i);
				}
			}
						
//inMsg -> PrintToStream();	// XXX ?!?!?!?! index is not as it should be according to the Book
//			int32 index;
//			for ( int i = 0;  inMsg -> FindInt32("index", i, &index) == B_OK;  i++ ) {
//cerr << index << endl;	// XXX
//				DebugStream *stream = assert_cast<StreamListItem *>(ItemAt(index)) -> Stream();
//				bool b = stream -> IsFileLoggingEnabled();
//				stream -> SetFile(0);				// 0 is the default file name
//				if (b != stream -> IsFileLoggingEnabled())
//					InvalidateItem(index);
//			}
			break;
		}
		
		case M_CLEAR_SELECTED_LOGS: {
			BAutolock l(gGetDebugManager().Locker());
			
			int count = CountItems();
			for ( int i = 0;  i < count;  i++ ) {
				if (IsItemSelected(i)) {
//					DebugStream *stream = assert_cast<StreamListItem *>(ItemAt(i)) -> Stream();
// +++ implement
//					bool b = stream -> IsFileLoggingEnabled();
//					stream -> SetFileLoggingEnabled(true);
//					if (b != stream -> IsFileLoggingEnabled())
//						InvalidateItem(i);
				}
			}
			break;
		}
			
		case M_LOG_TO_FILE: {
			BAutolock l(gGetDebugManager().Locker());
			
			void *ptr;
			if (inMsg -> FindPointer("sitem", &ptr) == B_OK) {
				StreamListItem *sitem = static_cast<StreamListItem *>(ptr);
				DebugStream *stream = sitem -> Stream();
				if (stream -> IsFileLoggingEnabled()) {
					stream -> SetFileLoggingEnabled(false);
				} else {
					stream -> SetFileLoggingEnabled(true);
				}
				InvalidateItem(IndexOf(sitem));
			}
			break;
		}
		
		case M_SELECT_ALL:
			Select(0, CountItems() - 1);
			break;
		case M_SELECT_NONE:
			DeselectAll();
			break;

		default:
			inherited :: MessageReceived(inMsg);
	}
}

void StreamListView :: MouseDown(BPoint inPoint) {
	int32 buttons = 0;
	
	Window() -> CurrentMessage() -> FindInt32("buttons", &buttons);
	if (buttons & B_SECONDARY_MOUSE_BUTTON) {
		
		StreamListItem *sitem = 0;
		for (int i = 0;  i < CountItems();  i++) {
			if (ItemFrame(i).Contains(inPoint)) {
				sitem = assert_cast<StreamListItem *>(ItemAt(i));
				break;
			}
		}
		
		BPopUpMenu	*menu;
		BMenuItem	*mitem;
		
		menu	= new BPopUpMenu("Options");
		
		if (sitem) {
			char buf[256];
			sprintf(buf, "Open window of '%s'", sitem -> Text());
			BMessage *msg = new BMessage(M_OPEN_WINDOW);
			msg		-> AddPointer("sitem", sitem);
			menu	-> AddItem(mitem = new BMenuItem(buf, msg, 'O'));
			mitem	-> SetTarget(this);
			
			sprintf(buf, "Log '%s' to file", sitem -> Text());
			msg = new BMessage(M_LOG_TO_FILE);
			msg		-> AddPointer("sitem", sitem);
			menu	-> AddItem(mitem = new BMenuItem(buf, msg, 'L'));
			mitem	-> SetTarget(this);
			if (sitem -> Stream() -> IsFileLoggingEnabled())
				mitem -> SetMarked(true);
			
			menu	-> AddSeparatorItem();
		}
		
		menu	-> AddItem(mitem = new BMenuItem("Log all selected to file", new BMessage(M_LOG_SELECTED_TO_FILE), 'S'));
		mitem	-> SetTarget(this);
		
		menu	-> AddItem(mitem = new BMenuItem("Select all", new BMessage(M_SELECT_ALL), 'A'));
		mitem	-> SetTarget(this);
		
		menu	-> AddItem(mitem = new BMenuItem("Select none", new BMessage(M_SELECT_NONE), 'N'));
		mitem	-> SetTarget(this);
		
		ConvertToScreen(&inPoint);
		menu -> Go(inPoint, true, false, true);
	} else {
		inherited :: MouseDown(inPoint);
	}
}

MainWindow :: MainWindow(DebugManager *inManager) : BWindow(BRect(0, 0, 100, 100), "", B_TITLED_WINDOW, 0),
													mManager( inManager ), mListView(0) {
	thread_info ti;
	app_info	ai;
					
	get_thread_info(find_thread(0), &ti);
	be_roster -> GetRunningAppInfo(ti.team, &ai);
	BString s(ai.ref.name);
	s += " debug manager";
	SetTitle(s.String());
	
	BMenuBar *bar = new BMenuBar(BRect(0, 0, 32767, 30), "menubar");
	AddChild(bar);
	bar -> ResizeToPreferred();
	BMenu *menu = new BMenu("Options");
	BMenuItem *item;
	menu -> AddItem(item = new BMenuItem("Clear log at startup", new BMessage(M_CLEAR_LOG_AT_STARTUP)));
	item -> SetMarked(gGetDebugManager().mClearLogsAtStartup);
	
	menu -> AddItem(item = new BMenuItem("Log to file by default", new BMessage(M_LOG_TO_FILE_BY_DEFAULT)));
	item -> SetMarked(gGetDebugManager().mLogToFileByDefault);
	
	bar -> AddItem(menu);
	
	mListView = new StreamListView( mManager );
	AddChild( mListView );
	BRect rect(0, 0, 100, 100);
	rect.top = bar -> Bounds().Height() + 1.;
	mListView -> MoveTo(rect.LeftTop());
	mListView -> ResizeTo(rect.Width(), rect.Height());
	
	Preferences *prefs = gGetUserPreferences();
	Autolock<Preferences> ll(prefs);
	
	if ( prefs -> FindRect("dstream:MainWindow.frame", &rect) == B_OK ) {
		MoveTo(rect.LeftTop());
		ResizeTo(rect.Width(), rect.Height());
	} else {
		MoveTo(50, 50);
		ResizeTo(150, 200);
	}
}

MainWindow :: ~MainWindow() {
	Preferences *prefs = gGetUserPreferences();
	Autolock<Preferences> ll(prefs);
	
	BRect rect = Frame();
	prefs -> SetRect("dstream:MainWindow.frame", rect);
}

int compare_func(const StreamListItem **first, const StreamListItem **second) {
	return ( strcmp((*first) -> Text(), (*second) -> Text()) );
}

void MainWindow :: MessageReceived(BMessage *inMsg) {

	switch ( inMsg -> what ) {
		case M_WINDOW_GONE:
			mListView -> Invalidate(mListView -> Bounds());
			break;
		case M_NEW_STREAM: {
			BAutolock l(gGetDebugManager().Locker());
			
			DebugStream *stream;
			if ( inMsg -> FindPointer("Stream", (void **)&stream) == B_OK ) {
				mListView -> AddItem( new StreamListItem(stream) );
				mListView -> SortItems((int (*)(const void *, const void *))&compare_func);
			}
			break;
		}
		
		case M_CLEAR_LOG_AT_STARTUP: {
			BMenuItem *item;
			inMsg -> FindPointer("source", (void **)&item);
			bool newstate =  ! item -> IsMarked();
			item -> SetMarked(newstate);
			gGetDebugManager().mClearLogsAtStartup = newstate;
			break;
		}
		case M_LOG_TO_FILE_BY_DEFAULT: {
			BMenuItem *item;
			inMsg -> FindPointer("source", (void **)&item);
			bool newstate =  ! item -> IsMarked();
			item -> SetMarked(newstate);
			gGetDebugManager().mLogToFileByDefault = newstate;
			break;
		}
		default:
			inherited :: MessageReceived(inMsg);
	}
}





#endif // USE_DEBUGSTREAM