#if !defined(DEBUGSTREAM_H)
#define DEBUGSTREAM_H

//	Copyright (C) by Attila Lendvai
//	You are free to use this file in anything you want as long as it's clear
//	that it's not your work. Use at your own risk, no warranties!

#define USE_DEBUGSTREAM 	1		// comment out to avoid all debugging code

#if USE_DEBUGSTREAM

#include <iostream>
#include <Window.h>
#include <AppDefs.h>
#include <Locker.h>
#include <Entry.h>
#include <ScrollView.h>
#include <TextView.h>
#include <ListView.h>
#include <ListItem.h>
#include <TLS.h>
#include <String.h>

#include <vector>

namespace debug {
	class DebugStream;
	class DebugManager;
}

namespace debug_internal {
	enum {
		M_NEW_STREAM = 1747698320,
		M_ITEMS_CLICKED,
		M_LOG_TO_FILE,
		M_LOG_SELECTED_TO_FILE,
		M_SELECT_ALL,
		M_SELECT_NONE,
		M_WINDOW_GONE,
		M_OPEN_WINDOW,
		M_CLEAR_LOG_AT_STARTUP,
		M_LOG_TO_FILE_BY_DEFAULT,
		M_CLEAR_SELECTED_LOGS
	};
	
	class TextView : public BTextView {
		typedef BTextView	inherited;
		bool				mNeedSetTextRect;
	public:
		TextView();
				
				void			Pulse();
				void			FrameResized(float, float);
	};
	
	
	class Window : public BWindow {
		friend class Buffer;
		
		TextView				*mTextView;
		BScrollView				*mScrollView;
		debug :: DebugStream	*mStream;
											
		static		int32			sInstances;
		static		BPoint			sLeftTop;
	public:
		Window();
		~Window();

				void			Detach(debug :: DebugStream *stream);
				void			Attach(debug :: DebugStream *stream); 
//				void			close();		
//				void			open( const char inName[] );
				void			Append(const char *, int);
//				void			FrameResized(float, float);
	};
	
	class StreamListItem : public BStringItem {
		debug :: DebugStream	*mStream;
	public:
		StreamListItem(debug :: DebugStream *);
		
			void					DrawItem(BView *, BRect, bool);
		debug :: DebugStream *	Stream() const	{ return mStream; }
	};

	class StreamListView : public BListView {
		typedef BListView inherited;
		debug :: DebugManager	*mManager;
	
	public:
		StreamListView(debug :: DebugManager *);
		
			void		AllAttached()				{ SetTarget(BMessenger(this, Looper())); }
			void		MessageReceived(BMessage *);
			void		MouseDown(BPoint);
	};

	class MainWindow : public BWindow {
		debug :: DebugManager	*mManager;
		StreamListView			*mListView;
		
	public:
		MainWindow(debug :: DebugManager *);
		~MainWindow();
		
			void		MessageReceived(BMessage *);
	};

	class Buffer : public filebuf { 
		friend class Window;
		
		typedef filebuf	inherited;
		debug :: DebugStream		*mOwner;
		BString						mFileName;
//		BString						mBuffer;
		char *						mPreviousFlushAt;
		bool							mFileLoggingState;
		bool							mFlushed;
													
				void					CreateTimeStamp(char *);
																		
	public:
		Buffer(debug :: DebugStream *);
//		~Buffer();
		
				bool			IsFileLoggingEnabled()		{ return mFileLoggingState; }
				void			SetFileLoggingEnabled(bool b);
				BString	*		FileName()					{ return &mFileName; }
				void			SetFile( const char *inName ) { mFileName = inName; }
				void			ReopenFile();
				void			CloseFile()					{ close(); }

		virtual	int			sync();
		virtual	int			doallocate();

//				void			close();
  		virtual	streamsize	xsputn(const char* s, streamsize n);
//		virtual	int			underflow();
		virtual	int			overflow(int ch);
//		virtual	int			doallocate();
//		virtual	streamsize	sys_read(char* buf, streamsize size);
//		virtual	streampos	sys_seek(streamoff, _seek_dir);
//		virtual	streamsize	sys_write(const char*, streamsize);
	};
}

namespace debug {
	namespace dbi = debug_internal;
	
	class DebugStream : public ostream {
		friend class dbi :: Buffer;
		thread_id					mThreadID;
		dbi :: Window				*mWindow;
		bool						mTimeStampState;
		dbi :: Buffer				mBuffer;
		BString						mName;
				
		static	 DebugStream			sGlobal;		
	public:
		DebugStream(thread_id id, const char *fname = (const char *)-1);
		DebugStream :: ~DebugStream();
										
			bool			IsFileLoggingEnabled()			{ return mBuffer.IsFileLoggingEnabled(); }
			void			SetFileLoggingEnabled(bool b)	{ mBuffer.SetFileLoggingEnabled(b); }
			void			SetFile(const char *inName) 	{ mBuffer.SetFile(inName); }
			BString	*		FileName()						{ return mBuffer.FileName(); }
			void			ReopenFile()					{ mBuffer.ReopenFile(); }
			thread_id		ThreadID() const 				{ return mThreadID; }
			bool			IsTimeStampEnabled()			{ return mTimeStampState; }
			const char *	Name();
			void			Attach(dbi :: Window *inWin);
			void			Detach(dbi :: Window *inWin);
			void			OpenWindow();
			dbi::Window* GetWindow()					{ return mWindow; }

			bool			HasWindow()		{ return (mWindow != 0); }
			void			ShowWindow()	{ if (mWindow) mWindow -> Show(); }
			void			HideWindow()	{ if (mWindow) mWindow -> Hide(); }
//			void			close()						{ mBuffer.close(); } 
//			void			AttachWindow(dbi :: Window *inWin)	{ mBuffer.AttachWindow(inWin); }
//			dbi::Window *Window()					{ return mBuffer.Window(); }
	};
									
	class DebugManager {
		friend dbi :: MainWindow;
		friend DebugStream;
		friend DebugStream	&gGetDebugStream();
		typedef vector<DebugStream *> stream_list_t;
		
				bool				mClearLogsAtStartup;
				bool				mLogToFileByDefault;
				entry_ref			mLogDir;
				stream_list_t		mStreams;
		mutable	BLocker				mLocker;
				dbi :: MainWindow	*mWindow;
				vint32				mInitialised;
		
		static	int32				sDebugStream_TLS;

				void				Initialise();
		
		DebugManager();
		~DebugManager();
		
	public:
		// Allow only one single global instance
		static	DebugManager	sManager;
											
		static	DebugManager &	Instance()				{ return sManager; }
				DebugStream	&	GetStream();
				stream_list_t *	StreamList()			{ return &mStreams; }
				BLocker *		Locker()				{ return &mLocker; }
				entry_ref *		LogDir()				{ Initialise(); return &mLogDir; }
				void			OpenMainWindowMember();
		static	void			OpenMainWindow();
				dbi::MainWindow *GetMainWindowMember()	{ return mWindow; }
		static	dbi::MainWindow *GetMainWindow()		{ return sManager.GetMainWindowMember(); }
	};
	
	inline DebugStream	&	gGetDebugStream()	{ return DebugManager :: Instance() . GetStream(); }
	inline DebugManager &	gGetDebugManager()	{ return DebugManager :: Instance(); }
}

#endif // USE_DEBUGSTREAM
#endif // DEBUGSTREAM_H
