/*
  TODO:
*/

#include "EditorBuffer.h"
#include "EditorView.h"
#include "ThreadMagic.h"

using namespace threadmagic;

#include <stdio.h>

EditorBuffer :: EditorBuffer() : fLocker("EditorBuffer :: fLocker") {
}

EditorBuffer :: ~EditorBuffer() {
	Lock();

	// Call Disconnected() on views and delete them
	size_t size = fViews.size();
	unsigned int i = 0;
	while (i < size) {
		fViews[i] -> Disconnected();
		delete fViews[i];
		fViews[i] = 0;
		i++;
	}
}

// Add inView to the list of views and call Connected()

void EditorBuffer :: ConnectView(EditorView *inView) {
	BAutolock l(Locker());
	if (l.IsLocked()) {
		fViews.push_back(inView);
		inView -> Connected(this);
	}
}

int32 EditorBuffer :: Data() {
	LockCheck();
	return fData;
}

// Find inView in the list of views and call Disconnected() and remove if found

bool EditorBuffer :: DisconnectView(EditorView *inView) {
	BAutolock l(Locker());
	if (l.IsLocked()) {
		view_vector_t :: iterator res = find(fViews.begin(), fViews.end(), inView);
		if (res == fViews.end()) {
			return false;
		} else {
			fViews.erase(res);
			(*res) -> Disconnected();
		}
	}
	return false;
}


// Here is the tricky part.

void EditorBuffer :: NotifyViews(FunctionObject *inFunctor) {
	LockCheck();	// Check if the buffer is locked !

	// Walk trough the list and call the given functions on them.
	view_vector_t :: iterator pos = fViews.begin(), end = fViews.end();
	while (pos < end) {
		inFunctor -> CallWithThis(*pos);	// Here *pos is a EditorView*, the current one in the list
		pos++;							// You must cast it to void * to call operator()(void *)
	}									// not operator()(AThread *). If you don't cast a warning
										// will tell you that some function is hidden.
	delete inFunctor;			// We are done, delete the functor

	puts("---");				// Just to clear the output
}


// This is a function to show how things work.
void EditorBuffer :: ChangeData(BString &inStr) {
	BAutolock l(Locker());
	if (l.IsLocked()) {
		fData = system_time();
		
		// Here is how to package a function into a FunctionObject and pass to NotifyViews()
		NotifyViews(new FunctionObjectTemplate
							<typeof(&EditorView :: FooBarNotification)>
							(&EditorView :: FooBarNotification, inStr)
							// Pass the string from the thread to the notification function
							// Note that the string will not be copied anywhere, only when 
							// the FunctionObject duplicates it. After that only a pointer
							// will be copied or passed. (If you don't want to duplicate
							// the string object then you should pass BString &* everywhere.
							// This way only a pointer is duplicated in the FunctionObject
							// ctor, but in this case you must take care that the passed object
							// is valid until the thread exists.)
					);
	}
}







