#include <iostream>
#include <exception>
#include <Roster.h>
#include <Entry.h>
#include <Path.h>
#include <OS.h>
#include <Debug.h>
#include <unistd.h>

#include "TestApp.h"
#include "DebugStream.h"
#include "ThreadMagic.h"
#include "Preferences.h"
#include "Misc.h"

using namespace std;
using namespace application;
using namespace misc;
using namespace debug;
using namespace threadmagic;
using namespace preferences;

const char *application :: signature = "application/x-vnd.101-DebugStream.testapp";

const bigtime_t default_delay = 3000000;

TestApp :: TestApp(const char *inSign) : BApplication(inSign) {
	thread_info ti;
	get_thread_info(find_thread(0), &ti);
	
	app_info ai;
	be_roster -> GetRunningAppInfo(ti.team, &ai);
	BPath path;
	BEntry appentry(&ai.ref), parent;
	appentry.GetParent(&parent);
	parent.GetPath(&path);
	parent.GetRef(&mAppHomeDir);
	chdir(path.Path());
}

void TestApp :: DispatchMessage(BMessage *inMsg, BHandler *inHandler) {
	try {
		inherited :: DispatchMessage(inMsg, inHandler);
	} catch (exception &e) {
		cerr << "An exception reached the top ('" << e.what()
			<< "'), it should not happen! Ignoring..." << endl;
	}
}

status_t TestApp :: SomeThread(const bigtime_t &inDelay) {
	DEBUG_ONLY(
		DebugStream &cdebug = gGetDebugStream();
		cdebug << "SomeThread started up. It will print something periodically.\n" << endl;
	)
	
	threadmagic::Thread *thread = Thread::ThisThread();
	
	while ( thread -> IsQuitRequested() == false ) {
//		cdebug << "Gonna write out 64k... ";	// R5 eats up all the memory after TextVeiw::Insert(, 60000);
//		for (int i = 0;  i < 0xffff; i++)
//			cdebug << "x";
//		
//		cdebug << " finished." << endl;
//		
//		cdebug << "After writing 64k..." << endl;

		DEBUG_ONLY(
			cdebug << "p";
			cdebug << "i";
			cdebug << "n";
			cdebug << "g" << endl;
		)
		snooze(inDelay);
	}
	return B_OK;
}

//void TestApp :: ReadyToRun() {
//}

int main() {
	try {
//		char *x = new char[100000000000];	// trigger an exception
		TestApp app(signature);
		
		cerr << "Starting TestCDebug... press CTRL-C if you are done!" << endl;
		#ifndef DEBUG
			cerr << "You've compiled the non-debug version. You'll see nothing here..." << endl
				<< "open the other project and compile that one!" << endl;
		#endif

		DEBUG_ONLY(
			DebugStream &cdebug = gGetDebugStream();
			cdebug << "This is the main thread speeking... startup completed." << endl;
		)
		
		Thread *thread = Thread :: Launch(new FunctionObjectTemplate<typeof(&TestApp::SomeThread)>(&app, &TestApp :: SomeThread, default_delay),
											"Pinger thread", B_NORMAL_PRIORITY);
							
		DEBUG_ONLY(
			cdebug << "Started pinger thread with delay: " << default_delay << endl;
	
			DebugManager :: OpenMainWindow();
		)
		
		app.Run();
		
		thread -> RequestToQuit(5000000, true);	// wait 5 secs then kill thread
		delete thread;
		
	} catch (exception &e) {
		cerr << "Some big problem happened right in the beginning! ('" << e.what()
		<< "') Exiting..." << endl;
	}
}


