//
//  Preferences.cpp V0.0.1 Copyright (C) 1998 by Attila Lendvai
//

#include <Path.h>
#include <FindDirectory.h>
#include <Directory.h>
#include <File.h>
#include <kernel/fs_attr.h>

#include <string.h>

#include "Preferences.h"

//
//	Implementation
//
Preferences :: Preferences(const char *inSignature, bool inCommon) : BMessage('pref') {
	initialise(inSignature, inCommon);
}

void Preferences :: initialise(const char *inSignature, bool inCommon) {
	fStatus		= B_NO_INIT;
	fCommon		= inCommon;

	char *start = (char *)inSignature;
	for (int i = strlen(inSignature) - 1;  i;  i--) {
		if (inSignature[i] == '/') {
			start = (char *) &inSignature[i + 1];
			break;
		}
	}
	char buf[B_MIME_TYPE_LENGTH + 32];
	strcpy(buf, start);
	strcat(buf, (fCommon) ? ".common" : ".user");

	BPath path;
	find_directory(B_COMMON_SETTINGS_DIRECTORY, &path, true);
	BDirectory dir(path.Path());

	BEntry entry(&dir, buf);
	entry.GetRef(&fEntryRef);
	if ((fStatus = entry.InitCheck()) == B_OK)
		Revert();
}

//Preferences :: ~Preferences() {
//}

status_t Preferences :: Revert() {
	status_t rc = B_NO_INIT;
	if (fStatus == B_OK) {
		BFile file(&fEntryRef, B_READ_ONLY);
		if ((rc = file.InitCheck()) == B_OK)
			rc = LoadFrom(&file);
	}
	return rc;
}

status_t Preferences :: LoadFrom(BFile *inFile) {
	status_t rc = B_NO_INIT;
	if (fStatus == B_OK) {
		for (int i = 0;  (rc = inFile -> Lock()) != B_OK  &&  i < 5;  i++)
			snooze(100000);
		if (rc == B_OK) {
			rc = Unflatten(inFile);
			inFile -> Unlock();
		}
	}
	return rc;
}

status_t Preferences :: SaveTo(BFile *inFile) {
	status_t rc = B_NO_INIT;
	if (fStatus == B_OK) {
		for (int i = 0;  (rc = inFile -> Lock()) != B_OK  &&  i < 5;  i++)
			snooze(100000);
		if (rc == B_OK) {
			rc = Flatten(inFile);
			inFile -> Unlock();
		}
	}
	return rc;
}

status_t Preferences :: Sync() {
	status_t rc = B_NO_INIT;
	if (fStatus == B_OK) {
		BEntry backupentry, entry(&fEntryRef);
		if (entry.Exists()) {
			backupentry = entry;
			char buf[B_FILE_NAME_LENGTH + 32];
			backupentry.GetName(buf);
			strcat(buf, ".backup");
			backupentry.Rename(buf, true);
		}
	
		BFile file(&entry, B_ERASE_FILE | B_CREATE_FILE | B_READ_WRITE);
	
		if ((rc = SaveTo(&file)) == B_OK) {
			backupentry.Remove();
		} else {
			file.Unset();
			char buf[B_FILE_NAME_LENGTH];
			entry.GetName(buf);
			if (backupentry.InitCheck() == B_OK)
				backupentry.Rename(buf, true);
		}
	}
	return rc;
}

status_t Preferences :: SetData(const char *inName, type_code inType, const void *inData, ssize_t inSize) {
	if (HasData(inName, inType))return ReplaceData(inName, inType, inData, inSize);
	else						return AddData(inName, inType, inData, inSize);
}

status_t Preferences :: SetPoint(const char *inName, BPoint inPoint) {
	if (HasPoint(inName))		return ReplacePoint(inName, inPoint);
	else						return AddPoint(inName, inPoint);
}

status_t Preferences :: SetRect(const char *inName, BRect inRect) {
	if (HasRect(inName))		return ReplaceRect(inName, inRect);
	else						return AddRect(inName, inRect);
}

status_t Preferences :: SetString(const char *inName, const char *inString) {
	if (HasString(inName))		return ReplaceString(inName, inString);
	else						return AddString(inName, inString);
}

status_t Preferences :: SetInt64(const char *inName, int64 inInt64) {
	if (HasInt64(inName))		return ReplaceInt64(inName, inInt64);
	else						return AddInt64(inName, inInt64);
}

status_t Preferences :: SetInt32(const char *inName, int32 inInt32) {
	if (HasInt32(inName))		return ReplaceInt32(inName, inInt32);
	else						return AddInt32(inName, inInt32);
}

status_t Preferences :: SetInt16(const char *inName, int16 inInt16) {
	if (HasInt16(inName))		return ReplaceInt16(inName, inInt16);
	else						return AddInt16(inName, inInt16);
}

status_t Preferences :: SetInt8(const char *inName, int8 inInt8) {
	if (HasInt8(inName))		return ReplaceInt8(inName, inInt8);
	else						return AddInt8(inName, inInt8);
}

status_t Preferences :: SetFloat(const char *inName, float inFloat) {
	if (HasFloat(inName))		return ReplaceFloat(inName, inFloat);
	else						return AddFloat(inName, inFloat);
}

status_t Preferences :: SetBool(const char *inName, bool inBool) {
	if (HasBool(inName))		return ReplaceBool(inName, inBool);
	else						return AddBool(inName, inBool);
}

status_t Preferences :: SetMessage(const char *inName, BMessage *inMessage) {
	if (HasMessage(inName))		return ReplaceMessage(inName, inMessage);
	else						return AddMessage(inName, inMessage);
}
