
#include <Alert.h>
#include <Bitmap.h>
#include <Mime.h>
#include <String.h>
#include <TranslatorRoster.h>

#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "DragView.h"

#define BM_WID 95
#define BM_HEI 31

using namespace threadmagic;

static void
make_image(
	BView * view)
{
	BRect r(view->Bounds());
	view->SetFont(be_bold_font);
	view->SetLowColor(255, 255, 255, 0);	//	transparent
	view->SetDrawingMode(B_OP_COPY);
	view->FillRect(r, B_SOLID_LOW);

	//	the colored squares are translucent when dragged
	BRect r2(r);
	r2.right = (r2.right+1)/3-1;
	r2.bottom = (r2.bottom+1)/2-1;
	view->SetHighColor(255,0,0, 128);
	view->FillRect(r2);

	r2.OffsetBy((r2.Width()+1)*2, 0);
	view->SetHighColor(0,255,0, 128);
	view->FillRect(r2);

	r2.OffsetBy(0,r2.Height()+1);
	view->SetHighColor(0,0,255, 128);
	view->FillRect(r2);

	r2.OffsetBy(-(r2.Width()+1)*2, 0);
	view->SetHighColor(255,0,255, 128);
	view->FillRect(r2);

	view->SetHighColor(64, 64, 64);
	view->SetFontSize(20);
	view->SetDrawingMode(B_OP_OVER);
	view->DrawString("Drag Me", BPoint((int)((BM_WID-view->StringWidth("Drag Me"))/2),24));
}

DragView::DragView(
	const BRect & area) :
	BView(area, "DragView", B_FOLLOW_ALL, B_WILL_DRAW), fDragRect(30, 70, 40, 80)
{
	BRect r(0,0,BM_WID,BM_HEI);
	m_bitmap = new BBitmap(r, B_RGB32, true);
	m_bitmapView = new BView(r, "_bmv", B_FOLLOW_NONE, B_WILL_DRAW);
	m_bitmap->Lock();
	m_bitmap->AddChild(m_bitmapView);
	make_image(m_bitmapView);
	m_bitmapView->Sync();
	m_bitmap->Unlock();
	m_mode = B_OP_ALPHA;	//	mode to drag bitmap with

	//	Find the translator to use for us
	translator_id * all_translators = NULL;
	int32 count = 0;
	status_t err = BTranslatorRoster::Default()->GetAllTranslators(&all_translators, &count);
	if (err >= B_OK) {
		err = B_ERROR;
		//	look through all available translators
		for (int ix=0; ix<count; ix++) {
			const translation_format * in_formats;
			int32 fmt_count;
			if (B_OK <= BTranslatorRoster::Default()->GetInputFormats(all_translators[ix],
					&in_formats, &fmt_count)) {
				//	look for translators that accept BBitmaps as input
				for (int iy=0; iy<fmt_count; iy++) {
					if (in_formats[iy].type == B_TRANSLATOR_BITMAP) {
						goto do_this_format;
					}
				}
			}
			continue;
do_this_format:
			const translation_format * out_formats;
			if (B_OK <= BTranslatorRoster::Default()->GetOutputFormats(all_translators[ix],
					&out_formats, &fmt_count)) {
				//	look through the output formats
				for (int iy=0; iy<fmt_count; iy++) {
					//	and take the first output format that isn't BBitmap
					if (out_formats[iy].type != B_TRANSLATOR_BITMAP) {
						m_translator = all_translators[ix];
						m_type_code = out_formats[iy].type;
						strcpy(m_the_type, out_formats[iy].MIME);
						err = B_OK;
						break;
					}
				}
			}
		}
	}
	delete[] all_translators;
	if (err < B_OK) {
		(new BAlert("", "There are no graphics-export translators installed.", "Stop"))->Go();
		Window()->PostMessage(B_QUIT_REQUESTED);
	}

	fPeriodicThread = 0;
	fThread = 0;
}

//status_t foobar(AThread *, int &, bool &, char &, int &) {
//	printf("foobar ok\n");
//	return B_OK;
//}

void DragView :: AllAttached() {
//	AThread :: Launch(new FunctionObject<void, typeof(&foobar), int, bool, char, int>(&foobar,
//												1, true, 'a', 2),
//								"foobar", 1);

	fPeriodicThread = PeriodicThread :: Launch(
								NewFunctionObject(this,
													&DragView :: PeriodicThreadFunc,
													BString("APeriodicThread param"),
													BPoint(20, 20)),
								300000, "DragView :: PeriodicThreadFunc", B_DISPLAY_PRIORITY);
	if (fPeriodicThread == 0)
		puts("Error launching APeriodicThread !");



	fCooperativeThread = Thread :: Launch(
								NewFunctionObject(this, &DragView :: CooperativeThreadFunc),
								"DragView :: CooperativeThreadFunc", 1);
	if (fCooperativeThread == 0)
		puts("Error launching AThread (cooperative) !");



	fThread = Thread :: Launch(
								NewFunctionObject(this, &DragView :: ThreadFunc, 12,
									"This is passed by reference, not duplicated !"),
									// We can do this because it is a constant string...
								"DragView :: ThreadFunc", 1);
	if (fThread == 0)
		puts("Error launching AThread (non-cooperative) !");
}

DragView::~DragView()
{
	if (fCooperativeThread -> State() > 0) {
		puts("CooperativeThreadFunc haven't exit yet ! Now I ask it to exit and wait 10 seconds...");
		fCooperativeThread -> RequestToQuit();
		if (fCooperativeThread -> WaitForThreadWithTimeout(10 * 1000000) == false) {
			puts("CooperativeThreadFunc haven't exited in time, killing it...\n");
			fCooperativeThread -> Kill();
			snooze(100000);	// wait for signal to get processed !
		}
	}
	delete fCooperativeThread;

	if (fThread -> State() > 0) {
		puts("ThreadFunc haven't exit yet ! Now I wait 10 seconds before killing it...");
		if (fThread -> WaitForThreadWithTimeout(10 * 1000000) == false) {
			puts("ThreadFunc haven't exited in time, killing it...\n");
			fThread -> Kill();
			snooze(100000);	// wait for signal to get processed !
		}
	}
	delete fThread;

	if (fPeriodicThread) {
		// *!*!* this timeout is way too big because sometimes the thread
		// get stuck and I'm checking the semaphores... so it's for debugging.
		//
		//  The problem is the following: DragWindow wants to quit, so it's
		//  locked while it's dtor gets called. And when we reach the DragView
		//  dtor we will wait here for a long time while the APeriodicThread
		//  will not exit because it's waiting for the DragWindow lock which
		//  will not return until we (the DragView dtor) don't return. Deadlock.
		//  I'm continously sending SIGCONT in RequestToQuit() but it does not
		//  seem to bother LockLooper() sometimes... (?)
		//  This deadlock is easily avoidable with a flag, but that seems to
		//  be a hack for me, and it's not important for an example...
		if (fPeriodicThread -> RequestToQuit(5000 * 1000000) == false) {
			puts("APeriodicThread didn't exited in time, killing it...");
			fPeriodicThread -> Kill();
		}
		delete fPeriodicThread;
	}

	delete m_bitmap;
}


// This is a method required for the chess thing...
bool hit(uint8 *x, int j ) {
	register int i = 0;
	while (i < j)  {
		if ( (x[j] == x[i]) || (abs(j - i) == abs(x[j] - x[i])) )
			return(true);
		i++;
	}
	return(false);
}

//	This thread will calculate that in how many variations you can place 8
//	queens on a 8x8 chess board without hitting each other. If it recognices
//	that it was asked to quit then it simply quits...
status_t DragView :: CooperativeThreadFunc() {

	Thread *thread = Thread::ThisThread();

	puts("CooperativeThreadFunc: started...");

	uint8 x[8];
	register int count = 0;
	int i;
	
	for (i = 0; i < 7; i++) x[i] = 0;
	
	for (; x[0] < 8  &&  ! thread -> IsQuitRequested(); x[0]++) {
		for (; x[1] < 8; x[1]++) {
			if (!hit(x, 1)) {
				for (; x[2] < 8; x[2]++) {
					if (!hit(x, 2)) {
						for (; x[3] < 8; x[3]++) {
							if (!hit(x, 3)) {
								for (; x[4] < 8; x[4]++) {
									if (!hit(x, 4)) {
										for (; x[5] < 8; x[5]++) {
											if (!hit(x, 5)) {
												for (; x[6] < 8; x[6]++) {
													if (!hit(x, 6)) {
														for (; x[7] < 8; x[7]++) {
															if (!hit(x, 7)) {
																count++;
							snooze(100000);	// slowdown...
															}
														}
														x[7]=0;
													}
												}
												x[6]=0;
											}
										}
										x[5]=0;
									}
								}
								x[4]=0;
							}
						}
						x[3]=0;
					}
				}
				x[2]=0;
			}
		}
		x[1]=0;
	}
	
	if (thread -> IsQuitRequested())
		puts("CooperativeThreadFunc: I was asked to quit...");
	else
		printf("CooperativeThreadFunc: You can place 8 queens on a chess board in %ld variations not hitting each other\n\n", (uint32)count>>2);

	return B_OK;
}

//	This thread will calculate the same but for arbitrary sized board. It also
//	receives a char * that's not duplicated. Take care of passing pointers !!!
//	Now we can do that because this is a constant string defined at the line
//	the thread is launched in AllAttached().
status_t DragView :: ThreadFunc(int &inSize, const char *&inStr) {

	printf("ThreadFunc: started with string '%s'\n\n", inStr);

	uint8 x[inSize];
	register int count = 0;
	register int j;

	int SIZE = inSize - 1;
	
	for (j = 0; j <= SIZE; j++) x[j] = 0;

	j = 0;
	do {
		while ((hit(x, j )) && (j >= 0)) {
			while ((++x[j] > SIZE) && (j >= 0)) {
				x[j] = 0;
				j--;
			}
		}
		
		if (j >= 0) {
			if (j >= SIZE) {
				count++;
				while ((++x[j] > SIZE) && (j >= 0)) {
					x[j] = 0;
					j--;
				}
			} else {
				j++;
			}
		}
	} while ((j >= 0) && (j <= SIZE));
	
	printf("ThreadFunc: You can place %ld queens on a %ldx%ld chess board in %ld variations not hitting each other\n\n",(uint32)inSize,(uint32)inSize,(uint32)inSize,(uint32)count>>2);
	
	return count >> 2;

//	srandom(real_time_clock());
//	while (1) {
//		int i = random();
//		if ((i & 0xffffff) == 0) {	// you can change the calculating time with this
//			printf("ThreadFunc: found a good random: %lx\n", (uint32)i);
//			return i;
//		}
//	}
}

//	This thread will periodically write the string in the window until it's asked
//	to quit or LockLooper() fails (the window is gone) (We could also check if the
//	thread was asked to quit before LockLooper() to avoid deadlock... Always do so
//	in real applications !)
bool DragView :: PeriodicThreadFunc(BString &inStr, BPoint &inPoint) {
	if (LockLooper()) {
		rgb_color color = {random() & 0xff, random() & 0xff, random() & 0xff, random() & 0xff};
		SetHighColor(color);
		color.red 	= uint8((float)color.red * 0.75);
		color.green = uint8((float)color.green * 0.75);
		color.blue 	= uint8((float)color.blue * 0.75);
		SetLowColor(color);
		DrawString(inStr.String(), inPoint);
		Sync();
		UnlockLooper();
	} else {
		puts("PeriodicThreadFunc: LockLooper() failed, exiting...");
		return false;	// This will cause the thread to exit...
	}
	
	return true;
}

//	This thread will be spawned when you click on the white box and start
//	dragging
bool DragView :: MouseTrackingFunc(BPoint &inPoint, uint32 inButtons) {
	BRect rect(fDragRect);
	fDragRect.Set(inPoint.x, inPoint.y, inPoint.x + 10, inPoint.y + 10);
	rect = rect | fDragRect;
	if (LockLooper()) {
		Draw(rect);
		UnlockLooper();
	}
	return true;
}

void
DragView::Draw(
	BRect area)
{
	SetLowColor(216, 216, 216);
	SetHighColor(0,0,0);
	SetDrawingMode(B_OP_COPY);
	//	figure out where bitmap goes
	BRect r(m_bitmap->Bounds());
	int zz = (int)(Bounds().right-r.right)/2;
	r.OffsetBy(zz, zz);
	DrawBitmapAsync(m_bitmap, r.LeftTop());
	//	Draw gray area outside of frame
	BRect r2 = Bounds();
	r2.bottom = r.top-1;
	FillRect(r2, B_SOLID_LOW);
	r2.top = r.bottom+1;
	r2.bottom = Bounds().bottom;
	FillRect(r2, B_SOLID_LOW);
	r2.bottom = r.bottom;
	r2.top = r.top;
	r2.right = r.left-1;
	FillRect(r2, B_SOLID_LOW);
	r2.left = r.right+1;
	r2.right = Bounds().right;
	FillRect(r2, B_SOLID_LOW);
	//	Draw frame
	r2 = r;
	r2.InsetBy(-1, -1);
	StrokeRect(r2);

	SetHighColor(255, 255, 255);
	StrokeRect(fDragRect);
}

void
DragView::MouseDown(
	BPoint where)
{
#if 0	// this is obsolete, the be api provides better a way to handle this.
		// use mousemoved messages instead
	if (fDragRect.Contains(where)) {
		puts("Starting CPPMouseDownThread");
		CPPMouseDownThread<DragView> :: Launch(this, &DragView :: MouseTrackingFunc, 0, 100000);
		return;
	}
#endif

	BRect r(m_bitmap->Bounds());
	int zz = (int)(Bounds().right-r.right)/2;
	r.OffsetBy(zz, zz);
	if (r.Contains(where)) {
		//	wait for a drag to start
		while (true) {
			BPoint w2;
			uint32 mods;
			GetMouse(&w2, &mods);
			if (!mods) {
				//	releasing the buttons without moving means no drag
				goto punt;
			}
			if (w2 != where) {
				//	moving the mouse starts a drag
				break;
			}
			snooze(40000);
		}
		BMessage msg(B_SIMPLE_DATA);
		//	add the types
		msg.AddString("be:types", B_FILE_MIME_TYPE);
		msg.AddString("be:types", m_the_type);
		msg.AddString("be:filetypes", m_the_type);
		//	add the actions
		msg.AddInt32("be:actions", B_COPY_TARGET);
		msg.AddInt32("be:actions", B_TRASH_TARGET);
		msg.AddString("be:clip_name", "DragMe Bitmap");
		//	create new bitmap to drag around
		BBitmap * bmp = new BBitmap(m_bitmap->Bounds(), m_bitmap->ColorSpace(), false);
		memcpy(bmp->Bits(), m_bitmap->Bits(), m_bitmap->BitsLength());
		//	start the drag (the rest will be handled by MessageReceived())
		BPoint pt(where.x-r.left, where.y-r.top);
		DragMessage(&msg, bmp, B_OP_ALPHA, pt, Window());
		return;
	}
punt:
	Window()->Activate(true);
}

